#!/bin/bash

usage()
{
    nom=`basename $0`
    echo "*****"
    echo "  $nom -a (32|64) -c (debug|release) -l /my/absolute/install/libs/path -o /my/absolute/install/path"
    echo "  $nom -h"
    echo "*****"
}

arch=0
configuration=""
libspath=""
outpath=""

# Cas particulier de l'option -h: doit être seule
if [[ $# -eq 0 ]]; then usage; exit 2; fi
if [[ $# -eq 1 && "$1" = "-h" ]]; then usage; exit 2; fi

#http://www.shellunix.com/ksh.html
while getopts ":a:c:l:o:" option
do
    case $option in
        a) arch=$OPTARG
        ;;
        :) usage
            exit 1
        ;;

        c) configuration=$OPTARG
        ;;
        :) usage
            exit 1
        ;;

        l) libspath=$OPTARG
        ;;
        :) usage
            exit 1
        ;;

        o) outpath=$OPTARG
        ;;
        :) usage
            exit 1
        ;;
        
        \?) usage
            exit 1
        ;;
    esac
done

if [[ $arch -ne 32 && $arch -ne 64 ]]; then
    usage
    exit 1
fi
if [[ x"$configuration" != x"debug" && x"$configuration" != x"release" ]]; then
    usage
    exit 1
fi
if [[ x"$libspath" = x ]]; then
    usage
    exit 1
fi
if [[ x"$outpath" = x ]]; then
    usage
    exit 1
fi

#---

arch_flag=-m$arch
debug_flags=
config=
if [[ x"$configuration" = x"debug" ]]; then
    debug_flags="-g -DDEBUG"
    config=d
fi

zlib_flags=
zlib_flags="${zlib_flags} -I$libspath/include"
zlib_flags="${zlib_flags} -L$libspath/lib -lz"

fix_flags=
if [[ $arch -eq 32 ]]; then
    fix_flags=" -read_only_relocs suppress"
fi

#---

mkdir -p $outpath

g++ $debug_flags -o $outpath/FFmpegLoader-$arch$config.so \
    -O3 -shared $arch_flag $fix_flags \
    -I$libspath/include \
    $zlib_flags \
    ../../src/LoaderFFmpeg.cpp \
    ../../src/Format.LoaderFFmpeg.cpp \
    $libspath/lib/libavformat.a \
    $libspath/lib/libavcodec.a \
    $libspath/lib/libavutil.a \
    $libspath/lib/libavdevice.a \
    $libspath/lib/libavfilter.a \
    $libspath/lib/libswscale.a \
    $libspath/lib/libswresample.a \
    $libspath/lib/libopenh264.a \
    $libspath/lib/libbz2.a \
    $libspath/lib/libvpx.a \
    $libspath/lib/libopus.a \
    -framework CoreFoundation -framework VideoDecodeAcceleration -framework CoreVideo -framework VideoToolbox -framework AudioToolbox -framework CoreMedia -framework Security -liconv -llzma -stdlib=libc++


g++ $debug_flags -o $outpath/FFmpegSaver-$arch$config.so \
    -O3 -shared $arch_flag $fix_flags \
    -I$libspath/include \
    $zlib_flags \
    ../../src/SaverFFmpeg.cpp \
    ../../src/Format.SaverFFmpeg.cpp \
    $libspath/lib/libavformat.a \
    $libspath/lib/libavcodec.a \
    $libspath/lib/libavutil.a \
    $libspath/lib/libavdevice.a \
    $libspath/lib/libavfilter.a \
    $libspath/lib/libswscale.a \
    $libspath/lib/libswresample.a \
    $libspath/lib/libopenh264.a \
    $libspath/lib/libbz2.a \
    $libspath/lib/libvpx.a \
    $libspath/lib/libopus.a \
    -framework CoreFoundation -framework VideoDecodeAcceleration -framework CoreVideo -framework VideoToolbox -framework AudioToolbox -framework CoreMedia -framework Security -liconv -llzma -stdlib=libc++

