/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE TVPAINT ANIMATION LOADER/SAVER PLUGINS  *
 * SOURCE CODE.  													*
 *																	*
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS LIBRARY SOURCE IS     *
 * GOVERNED BY A FREEBSD SOURCE LICENSE INCLUDED WITH THIS SOURCE   *
 * IN 'COPYING'. PLEASE READ THESE TERMS BEFORE DISTRIBUTING.       *
 *                                                                  *
 * THE TVPAINT ANIMATION LOADER/SAVER PLUGINS SOURCE CODE IS (C)    *
 * COPYRIGHT 2010-2015 BY THE TVPAINT DEVELOPPEMENT COMPANY         *
 * http://www.tvpaint.com/                                          *
 *                                                                  *
 ********************************************************************/

// ==================================================================
/** @file Format.BaseLoader.h
      @brief Header File: Format.BaseLoader
     @author Michael Schreiner & Herve ADAM
      $Date: $
        $Id: $
  @copyright (c) 2010-2015 TVPaint Developpement. All Rights Reserved.
*/
// ==================================================================

#ifndef __FORMAT_BASELOADER_HH
#define __FORMAT_BASELOADER_HH

#ifndef __FORMAT_BASE_HH
#include "Format.Base.h"
#endif

/**
* Soundtrack of the movie
*/
class cLoaderAudioSequence
{
public:
    virtual ~cLoaderAudioSequence()
    {
    }

    cLoaderAudioSequence():
        mChannelCount(0),
        mChannelType(kChannelTypesUnknown),
        mBitsPerChannel(0),
        mBytesPerSample(0),
        mSampleCount(0),
        mSamplingRate(0)
    {
    }
    cLoaderAudioSequence(int iChannelCount, eChannelTypes iChannelType, int iBitsPerChannel, int iBytesPerSample, int iSampleCount, int iSamplingRate):
        mChannelCount(iChannelCount),
        mChannelType(iChannelType),
        mBitsPerChannel(iBitsPerChannel),
        mBytesPerSample(iBytesPerSample),
        mSampleCount(iSampleCount),
        mSamplingRate(iSamplingRate)
    {
    }

public:
    /** Channel count getter & setter */
    int  ChannelCount() { return mChannelCount; }
    void ChannelCount( int iChannelCount ) { mChannelCount = iChannelCount; }
    /** Channel type getter & setter */
    eChannelTypes ChannelType() { return (eChannelTypes)mChannelType; }
    void ChannelType( eChannelTypes iChannelType ) { mChannelType = iChannelType; }
    /** Bit per channel getter & setter */
    int  BitsPerChannel() { return mBitsPerChannel; }
    void BitsPerChannel( int iBitsPerChannel ) { mBitsPerChannel = iBitsPerChannel; }
    /** Bytes per sample getter & setter */
    int  BytesPerSample() { return mBytesPerSample; }
    void BytesPerSample( int iBytesPerSample ) { mBytesPerSample = iBytesPerSample; }
    /** Start offset in seconds getter & setter */
    //double Offset() { return mOffset; }
    //void   Offset( double iOffset ) { mOffset = iOffset; }
    /** Sample count getter & setter */
    int  SampleCount() { return mSampleCount; }
    void SampleCount( int iSampleCount ) { mSampleCount = iSampleCount; }
    /** Sampling rate in Hz getter & setter */
    int  SamplingRate() { return mSamplingRate; }
    void SamplingRate( int iSamplingRate ) { mSamplingRate = iSamplingRate; }

protected:
    int mChannelCount;
    int/*eChannelTypes*/ mChannelType;
    int mBitsPerChannel;
    int mBytesPerSample;    // 16bits stereo (*2) => 4 bytes
    //double mOffset;
    int mSampleCount;
    int mSamplingRate;
};


/**
* Videotrack of the movie
*/
class cLoaderFrameSequence
{
public:
    virtual ~cLoaderFrameSequence()
    {
    }

    cLoaderFrameSequence():
        mWidth(0),
        mHeight(0),
        mFieldOrder(kNone),
        mFrameCount(0),
        mFrameRate(0),
        mPixelAspectRatio(0)
    {
    }

    cLoaderFrameSequence(int iWidth, int iHeight, eFieldOrder iFieldOrder, int iFrameCount, double iFrameRate, double iPixelAspectRatio):
        mWidth(iWidth),
        mHeight(iHeight),
        mFieldOrder(iFieldOrder),
        mFrameCount(iFrameCount),
        mFrameRate(iFrameRate),
        mPixelAspectRatio(iPixelAspectRatio)
    {
    }


public:
    /** Width getter & setter */
    int  Width() const { return  mWidth; }
    void Width( int iWidth ) { mWidth = iWidth; }
    /** Height getter & setter */
	int  Height() const { return  mHeight; }
	void Height( int iHeight ) { mHeight = iHeight; }
	/** Field order getter & setter */
	eFieldOrder FieldOrder() const { return  mFieldOrder; }
	void FieldOrder( eFieldOrder iFieldOrder ) { mFieldOrder = iFieldOrder; }
	/** Frame count getter & setter */
	int  FrameCount() const { return  mFrameCount; }
	void FrameCount( int64_t iFrameCount ) { mFrameCount = (int)iFrameCount; }
	/** Frame rate getter & setter */
	double FrameRate() const { return  mFrameRate; }
	void   FrameRate(double iFrameRate ) { mFrameRate = iFrameRate; }
	/** Pixel aspect ratio getter & setter */
	double PixelAspectRatio() const { return  mPixelAspectRatio; }
	void   PixelAspectRatio( double iPixelAspectRatio ) { mPixelAspectRatio = iPixelAspectRatio; }

public:
	int     		mWidth;
    int     		mHeight;
	eFieldOrder  	mFieldOrder;
	int     		mFrameCount;
	double  		mFrameRate;
	double  		mPixelAspectRatio;
};


/**
* Output movie format
*/
class cLoaderFormat
{
public:
    virtual ~cLoaderFormat() = 0;

protected:
    cLoaderFormat()
    {
    }

public:
	virtual bool Open( const char* iFilePath ) = 0;
	virtual void Close() = 0;

	virtual const char* Version() = 0;

    virtual bool GetInfos( cLoaderFrameSequence* oFrameSequence ) = 0;
	virtual bool GetInfos( cLoaderAudioSequence* oAudioSequence ) = 0;
	virtual bool GetInfos( cLoaderAudioSequence* oAudioSequence, int iPreferedChannelsCount, int iPreferedBitsPerChannel, int iPreferedSamplingRate ) = 0;

    virtual bool ReadFrame( int iFrameIndex, cPixelBlockFF* oDst ) = 0;

    virtual bool ReadAudio( cAudioBlock* oDst ) = 0;

};


/** The types of the class factories */
typedef cLoaderFormat* (*tCreateLoaderFormat)();


#endif // __FORMAT_BASELOADER_HH
