/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE TVPAINT ANIMATION LOADER/SAVER PLUGINS  *
 * SOURCE CODE.                                                     *
 *                                                                  *
 * This file uses some components of the FFmpeg project.            *
 * FFmpeg is distributed under the terms of the GNU Lesser General  *
 * Public License as published by the Free Software Foundation.     *
 *                                                                  *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS LIBRARY SOURCE IS     *
 * GOVERNED BY THE LGPL LICENSE. PLEASE READ THESE TERMS BEFORE     *
 * DISTRIBUTING.                                                    *
 *                                                                  *
 * THE TVPAINT ANIMATION LOADER/SAVER PLUGINS SOURCE CODE IS (C)    *
 * COPYRIGHT 2010-2015 BY THE TVPAINT DEVELOPPEMENT COMPANY         *
 * http://www.tvpaint.com/                                          *
 *                                                                  *
 ********************************************************************/

// ==================================================================
/** @file Format.LoaderFFmpeg.cpp
      @brief Header File: Format.LoaderFFmpeg
     @author Michael Schreiner
      $Date: $
        $Id: $
  @copyright (c) 2010-2015 TVPaint Developpement. All Rights Reserved.
*/
// ==================================================================

#ifndef __FORMAT_LOADERFFMPEG_HH
#include "Format.LoaderFFmpeg.h"
#endif

#include <stdio.h>
#include <limits.h>

//---

cLoaderFormat::~cLoaderFormat()
{

}

cLoaderFormatFFmpeg::~cLoaderFormatFFmpeg()
{
	Close();
}

cLoaderFormatFFmpeg::cLoaderFormatFFmpeg() :
	cLoaderFormat(),

    mFormatContextPtr( 0 ),

    mVideoStreamId( -1 ),
	mVideoStream( 0 ),
    mAudioStreamId( -1 ),
	mAudioStream( 0 ),

    mVideoHeader(),
	mVideoCodecContextPtr( 0 ),
    mVideoCodecPtr( 0 ),
	mVideoStartPTS( 0 ),
	mVideoSwsContext( 0 ),

	mAudioHeader(),
	mAudioCodecContextPtr( 0 ),
	mAudioCodecPtr( 0 ),

	mAudioSwrContext( 0 ),
    mAudioBufferFrame( 0 ),

	mAudioDstSamplesCountMax( 0 ),
	mAudioDstSamplesData( 0 ),
	mAudioDstSamplesLinesize( 0 )
{
	av_register_all();

#ifdef DEBUG
	//av_log_set_level( AV_LOG_TRACE );
	//av_log_set_level( AV_LOG_VERBOSE );
#endif
}


bool
cLoaderFormatFFmpeg::GetInfos( cLoaderFrameSequence* oFrameSequence )
{
	if( !oFrameSequence )
		return false;

	oFrameSequence->Width( Width() );
	oFrameSequence->Height( Height() );
	oFrameSequence->FieldOrder( FieldOrder() );
	oFrameSequence->FrameCount( ImageCount() );
	oFrameSequence->FrameRate( av_q2d( FrameRate() ) );
	oFrameSequence->PixelAspectRatio( AspectRatio() );		//AspectRatio() du pixel ?? sûr ??

	return true;
}

bool
cLoaderFormatFFmpeg::GetInfos( cLoaderAudioSequence* oAudioSequence )
{
	if( !oAudioSequence )
		return false;

	oAudioSequence->BitsPerChannel( BitsPerChannel() );
	oAudioSequence->BytesPerSample( BytesPerSample() );
	oAudioSequence->ChannelCount( ChannelCount() );
	oAudioSequence->ChannelType( ChannelType() );
	//	oAudioSequence->Offset( Offset() );
	//	oAudioSequence->SampleCount( SampleCount() );
	oAudioSequence->SamplingRate( SamplingRate() );

	return true;
}

bool
cLoaderFormatFFmpeg::GetInfos( cLoaderAudioSequence* oAudioSequence, int iPreferedChannelsCount, int iPreferedBitsPerChannel, int iPreferedSamplingRate )
{
	if( !oAudioSequence )
		return false;

	mAudioHeader.mChannelCount = iPreferedChannelsCount;
	mAudioHeader.mSamplingRate = iPreferedSamplingRate;
	mAudioHeader.mChannelType = mAudioHeader.mChannelCount == 1 ? kChannelTypesMono : kChannelTypesStereo;
	mAudioHeader.mBitsPerChannel = iPreferedBitsPerChannel;
	mAudioHeader.mBytesPerSample = mAudioHeader.mBitsPerChannel / 8 * mAudioHeader.mChannelCount;
	//mAudioHeader.mOffset         =
	//mAudioHeader.mSampleCount    =

	oAudioSequence->BitsPerChannel( BitsPerChannel() );
	oAudioSequence->BytesPerSample( BytesPerSample() );
	oAudioSequence->ChannelCount( ChannelCount() );
	oAudioSequence->ChannelType( ChannelType() );
	//	oAudioSequence->Offset( Offset() );
	//	oAudioSequence->SampleCount( SampleCount() );
	oAudioSequence->SamplingRate( SamplingRate() );

	return true;
}

bool
cLoaderFormatFFmpeg::ReadFrame( int iFrameIndex, cPixelBlockFF* oDst )
{
	if( !mVideoStream )
		return false;

	return LoadVideoByIndexSeeking( oDst, iFrameIndex );
}

bool
cLoaderFormatFFmpeg::ReadAudio( cAudioBlock* oDst )
{
	if( !mAudioStream )
		return false;

    return LoadAudio( oDst );
}

const char*
cLoaderFormatFFmpeg::Version()
{
	return LOADERSAVER_VERSION_STR " ( " LIBAVFORMAT_IDENT " / " LIBAVCODEC_IDENT " )";
}

//====================================================================================

void
cLoaderFormatFFmpeg::Close()
{
	//VIDEO
	if( mVideoStream )
	{
		if( mVideoCodecContextPtr )
		{
			avcodec_close( mVideoCodecContextPtr );
			mVideoCodecContextPtr = 0;
		}
	}

	//AUDIO
	if( mAudioStream )
	{
		av_frame_free( &mAudioBufferFrame );

		if( mAudioCodecContextPtr )
		{
			avcodec_close( mAudioCodecContextPtr );
			mAudioCodecContextPtr = 0;
		}
	}

	//FORMAT
	avformat_close_input( &mFormatContextPtr );
	mVideoStreamId = -1;
	mVideoStream = 0;
	mAudioStreamId = -1;
	mAudioStream = 0;
}

bool
cLoaderFormatFFmpeg::Open( const char* iFilePath )
{
    Close();

    //format
	int err = avformat_open_input( &mFormatContextPtr, iFilePath, NULL, NULL );
    if( err < 0 )
        return false;

    if( avformat_find_stream_info( mFormatContextPtr, NULL ) < 0 )
    {
		avformat_close_input( &mFormatContextPtr );
        return false;
    }

    //mFormatContextPtr->flags   |= AVFMT_FLAG_IGNIDX | AVFMT_FLAG_GENPTS;
    mFormatContextPtr->flags |= AVFMT_FLAG_GENPTS;

	//---

    bool video = OpenVideo();
    bool audio = OpenAudio();

    if( !(audio || video) )
    {
		Close();
        return false;
    }

	return true;
}


int64_t
cLoaderFormatFFmpeg::FrameToPts( int iFrame ) const
{
	AVRational q = av_mul_q( av_inv_q( mVideoHeader.mFrameRate ), av_inv_q( mVideoStream->time_base ) );
	return mVideoStartPTS + ( int64_t( iFrame ) * q.num ) / q.den;
}

int
cLoaderFormatFFmpeg::PtsToFrame( int64_t iPts ) const
{
	AVRational q = av_mul_q( mVideoHeader.mFrameRate, mVideoStream->time_base );
	return ( ( iPts - mVideoStartPTS ) * q.num ) / q.den;
}

int
cLoaderFormatFFmpeg::GetCodecDelay() const
{
	return ( ( mVideoCodecPtr->capabilities & AV_CODEC_CAP_DELAY ) ? mVideoCodecContextPtr->delay : 0 ) + mVideoCodecContextPtr->has_b_frames + 1 /* To be sure */;
}

bool
cLoaderFormatFFmpeg::OpenVideo()
{
	mVideoStreamId = av_find_best_stream( mFormatContextPtr, AVMEDIA_TYPE_VIDEO, -1, -1, 0, 0 );
	if( mVideoStreamId < 0 )
		return false;

	mVideoStream = mFormatContextPtr->streams[mVideoStreamId];
	mVideoCodecContextPtr = mVideoStream->codec;

	//CODEC
	mVideoCodecPtr = avcodec_find_decoder( mVideoCodecContextPtr->codec_id );
	if( !mVideoCodecPtr )
		return false;

#ifdef USE_HWACCEL_IN_THE_FUTURE
	AVCodec* hwaccel_codec = NULL;
	if( !strcmp( mVideoCodecPtr->name, "h264" ) ||
		!strcmp( mVideoCodecPtr->name, "hevc" ) ||
		!strcmp( mVideoCodecPtr->name, "mpeg4" )||
		!strcmp( mVideoCodecPtr->name, "vc1" )  ||
		!strcmp( mVideoCodecPtr->name, "vp9" ) )
	{
		// use FFmpeg hwaccels if possible
		std::string codecName = mVideoCodecPtr->name;
#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__) || defined(__NT__)
		std::string dxCodecName = codecName+"_dxva2";
		hwaccel_codec = avcodec_find_decoder_by_name( dxCodecName.c_str() );
#elif __APPLE__
		std::string vdaCodecName = codecName+"_vda";
		hwaccel_codec = avcodec_find_decoder_by_name( vdaCodecName.c_str() );
		if( !hwaccel_codec )
		{
			std::string vtbCodecName = codecName+"_videotoolbox";
			hwaccel_codec = avcodec_find_decoder_by_name( vtbCodecName.c_str() );
		}
#elif __linux__
		std::string vdpauCodecName = codecName+"_vdpau";
		hwaccel_codec = avcodec_find_decoder_by_name( vdpauCodecName.c_str() );
#endif
	}

	if( (!hwaccel_codec) || ( avcodec_open2( mVideoCodecContextPtr, hwaccel_codec, NULL ) < 0 ))
#endif
	if( avcodec_open2( mVideoCodecContextPtr, mVideoCodecPtr, NULL ) < 0 )
		return false;

	//HEADER PARAMETERS
	//hack for wrong codec timebase
	//if( mVideoCodecContextPtr->time_base.num > 1000 && mVideoCodecContextPtr->time_base.den == 1 )
	//	mVideoCodecContextPtr->time_base.den=1000;

	mVideoHeader.mVideoHeight = mVideoCodecContextPtr->height;
	mVideoHeader.mVideoWidth = mVideoCodecContextPtr->width;
	if( mVideoHeader.mVideoWidth == 0 || mVideoHeader.mVideoHeight == 0 )
		return false;
	mVideoHeader.mFrameRate = av_stream_get_r_frame_rate( mVideoStream );

	//---

	mVideoHeader.mAspectRatio = 1.0;
	if( mVideoStream->sample_aspect_ratio.num > 0 )
		mVideoHeader.mAspectRatio = 1.0 / av_q2d( mVideoStream->sample_aspect_ratio );
	else if( mVideoCodecContextPtr->sample_aspect_ratio.num > 0 )
		mVideoHeader.mAspectRatio = 1.0 / av_q2d( mVideoCodecContextPtr->sample_aspect_ratio );

	//---

	mVideoStartPTS = mVideoStream->start_time;
	if( mVideoStartPTS == AV_NOPTS_VALUE )
	{
		avcodec_flush_buffers( mVideoCodecContextPtr );

		if( av_seek_frame( mFormatContextPtr, mVideoStreamId, 0, 0 ) >= 0 )
		{
			AVPacket packet = { 0 };
			av_init_packet( &packet );
			do
			{
				if( av_read_frame( mFormatContextPtr, &packet ) < 0 )
					break;

				if( packet.stream_index == mVideoStreamId )
					mVideoStartPTS = packet.pts;

				av_packet_unref( &packet );
			} while( mVideoStartPTS == AV_NOPTS_VALUE );
		}
	}
	if( mVideoStartPTS == AV_NOPTS_VALUE )
		mVideoStartPTS = 0;

	// Read the first images to find the start pts
	/*
	mVideoStartPTS = ( AV_NOPTS_VALUE + 1 ) * -1;
	avcodec_flush_buffers( mVideoCodecContextPtr );
	if( av_seek_frame( mFormatContextPtr, mVideoStreamId, 0, 0 ) >= 0 )
	{
		for( int i = 0; i < 10; )
		{
			AVPacket packet = { 0 };
			av_init_packet( &packet );

			if( av_read_frame( mFormatContextPtr, &packet ) < 0 )
				break;

			if( packet.stream_index == mVideoStreamId && packet.pts != AV_NOPTS_VALUE )
			{
				i++;
				mVideoStartPTS = FFMIN( mVideoStartPTS, packet.pts );
				// Check if pts exists
			}

			av_packet_unref( &packet );
		}
	}
	if( mVideoStartPTS == AV_NOPTS_VALUE )
		mVideoStartPTS = 0;
	*/

	//---

	mVideoHeader.mImageCount = int( mVideoStream->nb_frames );
	if( !mVideoHeader.mImageCount && mVideoStream->duration != AV_NOPTS_VALUE )
	{
		AVRational q = av_mul_q( mVideoStream->time_base, mVideoHeader.mFrameRate );
		mVideoHeader.mImageCount = int( ( mVideoStream->duration * q.num ) / q.den );
	}
	if( !mVideoHeader.mImageCount )
	{
		int64_t stop_pts = mVideoStartPTS;
		avcodec_flush_buffers( mVideoCodecContextPtr );
		av_seek_frame( mFormatContextPtr, mVideoStreamId, FrameToPts( 1 << 29 ), AVSEEK_FLAG_BACKWARD );
		AVPacket packet = { 0 };
		av_init_packet( &packet );

		while( av_read_frame( mFormatContextPtr, &packet ) >= 0 )
		{
			if( packet.stream_index == mVideoStreamId && packet.pts != AV_NOPTS_VALUE && packet.pts > stop_pts )
				stop_pts = packet.pts;

			av_packet_unref( &packet );
		}

		mVideoHeader.mImageCount = PtsToFrame( stop_pts ) + 1;
	}

	if( !mVideoHeader.mImageCount )
		return false;

	//---

	mVideoHeader.mFieldOrder = kNone;
	//TODO: 2.8
	/*
	if( mVideoCodecContextPtr->coded_frame )
	{
		if( mVideoCodecContextPtr->coded_frame->top_field_first == 0 )
			mVideoHeader.mFieldOrder = kLowerFirst;
		else
			mVideoHeader.mFieldOrder = kUpperFirst;
	}
	*/

	//BUFFER
	return true;
}


bool
cLoaderFormatFFmpeg::OpenAudio()
{
	mAudioStreamId = av_find_best_stream( mFormatContextPtr, AVMEDIA_TYPE_AUDIO, -1, -1, 0, 0 );
	if( mAudioStreamId < 0 )
		return false;

	mAudioStream = mFormatContextPtr->streams[mAudioStreamId];
	mAudioCodecContextPtr = mAudioStream->codec;

	//HEADER PARAMETERS
	mAudioHeader.mChannelCount		= mAudioCodecContextPtr->channels;
	mAudioHeader.mSamplingRate		= mAudioCodecContextPtr->sample_rate;
	mAudioHeader.mChannelType		= mAudioHeader.mChannelCount == 1 ? kChannelTypesMono : kChannelTypesStereo;
	mAudioHeader.mBitsPerChannel	= av_get_bytes_per_sample( mAudioCodecContextPtr->sample_fmt ) * 8;
	mAudioHeader.mBytesPerSample	= av_get_bytes_per_sample( mAudioCodecContextPtr->sample_fmt ) * mAudioHeader.mChannelCount;
	//mAudioHeader.mOffset         =
	//mAudioHeader.mSampleCount    =

	//CODEC
	mAudioCodecPtr = avcodec_find_decoder( mAudioCodecContextPtr->codec_id );
	if( !mAudioCodecPtr )
		return false;

    if( avcodec_open2( mAudioCodecContextPtr, mAudioCodecPtr, NULL ) < 0 )
        return false;

	//BUFFER
	mAudioBufferFrame = av_frame_alloc();
    if( !mAudioBufferFrame )
    	return false;
	else
		return true;
}

int
cLoaderFormatFFmpeg::DecodeAudioPacket( AVPacket* oPacket, int* oGotFrame, cAudioBlock* oDst )
{
	*oGotFrame = 0;
	int decoded = oPacket->size;

	if( oPacket->stream_index != mAudioStreamId )
		return decoded;

	int ret = avcodec_decode_audio4( mAudioStream->codec, mAudioBufferFrame, oGotFrame, oPacket );
	if( ret < 0 )
		return ret;

	decoded = FFMIN( ret, oPacket->size );

	if( !*oGotFrame )
		return decoded;

	int src_channels = mAudioBufferFrame->channels;
	int src_sample_rate = mAudioBufferFrame->sample_rate;
	AVSampleFormat src_format = AVSampleFormat( mAudioBufferFrame->format );
	int dst_channels = ChannelCount();
	int dst_sample_rate = SamplingRate();
	AVSampleFormat dst_format = AV_SAMPLE_FMT_S16;	// should be a switch of BitsPerChannel()

	av_opt_set_int( mAudioSwrContext, "in_channel_count", src_channels, 0 );
	av_opt_set_int( mAudioSwrContext, "in_sample_rate", src_sample_rate, 0 );
	av_opt_set_sample_fmt( mAudioSwrContext, "in_sample_fmt", src_format, 0 );
	av_opt_set_int( mAudioSwrContext, "out_channel_count", dst_channels, 0 );
	av_opt_set_int( mAudioSwrContext, "out_sample_rate", dst_sample_rate, 0 );
	av_opt_set_sample_fmt( mAudioSwrContext, "out_sample_fmt", dst_format, 0 );

	int error = swr_init( mAudioSwrContext );
	if( error < 0 )
		return error;

	//---

	int src_sample_count = mAudioBufferFrame->nb_samples;

	int dst_samples_count = av_rescale_rnd( swr_get_delay( mAudioSwrContext, src_sample_rate ) + src_sample_count, dst_sample_rate, src_sample_rate, AV_ROUND_UP );
	if( mAudioDstSamplesCountMax == 0 )
	{
		mAudioDstSamplesCountMax = dst_samples_count;
		error = av_samples_alloc_array_and_samples( &mAudioDstSamplesData, &mAudioDstSamplesLinesize, dst_channels, dst_samples_count, dst_format, 0 );
		if( error < 0 )
			return -1;
	}
	if( dst_samples_count > mAudioDstSamplesCountMax )
	{
		av_freep( &mAudioDstSamplesData[0] );
		error = av_samples_alloc( mAudioDstSamplesData, &mAudioDstSamplesLinesize, dst_channels, dst_samples_count, dst_format, 1 );
		if( error < 0 )
			return -1;

		mAudioDstSamplesCountMax = dst_samples_count;
	}

	int nb_samples = swr_convert( mAudioSwrContext, mAudioDstSamplesData, mAudioDstSamplesCountMax, (const uint8_t **)mAudioBufferFrame->extended_data, src_sample_count );
	if( nb_samples < 0 )
		return -1;

	oDst->WriteSamples( mAudioDstSamplesData[0], nb_samples );

	// FLUSH remaining samples
	while( ( nb_samples = swr_convert( mAudioSwrContext, mAudioDstSamplesData, mAudioDstSamplesCountMax, 0, 0 ) ) > 0 )
		oDst->WriteSamples( mAudioDstSamplesData[0], nb_samples );

	if( nb_samples < 0 )
		return -1;

	return decoded;
}

bool
cLoaderFormatFFmpeg::LoadAudio( cAudioBlock* oDst )
{
	mAudioSwrContext = swr_alloc();
	if( !mAudioSwrContext )
		return false;

	//---

	//PACKET

	AVPacket         audioPacket;
	av_init_packet( &audioPacket );
	audioPacket.data = 0;
	audioPacket.size = 0;
	int got_frame = 0;
	while( av_read_frame( mFormatContextPtr, &audioPacket ) >= 0 )
	{
		// Must be set, because pointer in audioPacket are modified
		AVPacket packet_process = audioPacket;
		do
		{
			int ret = DecodeAudioPacket( &packet_process, &got_frame, oDst );
			if( ret < 0 )
				break;
			packet_process.data += ret;
			packet_process.size -= ret;
		} while( packet_process.size > 0 );
		av_packet_unref( &audioPacket );
	}

	audioPacket.data = 0;
	audioPacket.size = 0;
	do
	{
		DecodeAudioPacket( &audioPacket, &got_frame, oDst );
	} while( got_frame );

	//---

    if( mAudioDstSamplesData )
        av_free( mAudioDstSamplesData[0] );
	av_free( mAudioDstSamplesData );
	mAudioDstSamplesData = 0;

	mAudioDstSamplesCountMax = 0;

	swr_free( &mAudioSwrContext );

    return true;
}

//-------------------------------------------------------------------------------------

bool
cLoaderFormatFFmpeg::LoadVideoByIndexSeeking( cPixelBlockFF* oDst, int iIdx )
{
   return LoadVideo(oDst, iIdx);
}


bool
cLoaderFormatFFmpeg::LoadVideoByTimeSeeking( cPixelBlockFF* oDst, double iTime )
{
   return LoadVideo(oDst, -1, iTime);
}

bool
cLoaderFormatFFmpeg::LoadVideo( cPixelBlockFF* oDst, int iIdx, double iTime )
{
	mVideoSwsContext = sws_getContext( mVideoStream->codec->width, mVideoStream->codec->height, mVideoStream->codec->pix_fmt,
										oDst->Width(), oDst->Height(), AV_PIX_FMT_BGRA,
										SWS_BICUBIC, NULL, NULL, NULL );
	if( !mVideoSwsContext )
		return false;

	int64_t pts = FrameToPts( FFMAX( 0, iIdx - FFMAX( 15, GetCodecDelay() ) /* arbitrary minimum (>12). to be sure to decode the wanted image */ ) );
	int error = av_seek_frame( mFormatContextPtr, mVideoStreamId, pts, AVSEEK_FLAG_BACKWARD );
	if( error < 0 )
		return false;

	// I don't know why the below code doesn't work
	// It would improve speed to not load every times the previous 25 frames for codec which doesn't need it
	/*
	avcodec_flush_buffers( mVideoCodecContextPtr );

	av_init_packet( &mVideoPacket );
	mVideoPacket.data = 0;
	mVideoPacket.size = 0;

	bool found = false;
	while( av_read_frame( mFormatContextPtr, &mVideoPacket ) >= 0 )
	{
		if( mVideoPacket.stream_index == mVideoStreamId )
		{
			found = true;
			break;
		}
		av_packet_unref( &mVideoPacket );
	}
	if( !found )
		return false;

	if( 1)//mVideoPacket.pts > pts )
	//if( mVideoPacket.dts == pts && mVideoPacket.pts != mVideoPacket.dts )
	{
		pts = mVideoPacket.dts - 15 * mVideoPacket.duration;
		if( pts < mVideoStartPTS )
			pts = 0;

		av_packet_unref( &mVideoPacket );
		FlushRemaningFrame();

		int error = av_seek_frame( mFormatContextPtr, mVideoStreamId, pts, AVSEEK_FLAG_BACKWARD );
		if( error < 0 )
			return false;
	}
	else
	{
		av_packet_unref( &mVideoPacket );
		FlushRemaningFrame();

		int error = av_seek_frame( mFormatContextPtr, mVideoStreamId, pts, AVSEEK_FLAG_BACKWARD );
		if( error < 0 )
			return false;
	}
	*/
	avcodec_flush_buffers( mVideoCodecContextPtr );

	int frame_decoded = -10;
	//PACKET
	AVPacket         videoPacket;
	av_init_packet( &videoPacket );
	videoPacket.data = 0;
	videoPacket.size = 0;
	while( av_read_frame( mFormatContextPtr, &videoPacket ) >= 0 )
	{
		if( videoPacket.stream_index == mVideoStreamId && frame_decoded == -10 )
		{
			frame_decoded = PtsToFrame( videoPacket.pts ) - 1;
		}
		int got_frame = 0;
		// Must be set, because pointer in mVideoPacket are modified
		AVPacket process_packet = videoPacket;
		do
		{
			int ret = DecodeVideoPacket( &process_packet, &got_frame, &frame_decoded, oDst );
			if( ret < 0 )
				break;

			process_packet.data += ret;
			process_packet.size -= ret;
			if( got_frame && frame_decoded >= iIdx )
				break;

		} while( process_packet.size > 0 );
		av_packet_unref( &videoPacket );

		if( got_frame && frame_decoded >= iIdx )
			break;
	}

	if( frame_decoded < iIdx )	// To be sure to get the last frame(s) because at the end of the file (with some codec?) the above 'while' can reach the end of file without having 'got_frame'
	{
		// flush cached frames
		videoPacket.data = NULL;
		videoPacket.size = 0;
		int got_frame = 0;
		do
		{
			int ret = DecodeVideoPacket( &videoPacket, &got_frame, &frame_decoded, oDst );
			if( ret < 0 )
				break;

			if( got_frame && frame_decoded >= iIdx )
				break;
		} while( got_frame );
	}

	FlushRemaningFrame( &videoPacket );

	//av_free_packet( &mVideoPacket );

	return frame_decoded == iIdx;

	//---

	sws_freeContext( mVideoSwsContext );

	return true;
}

void
cLoaderFormatFFmpeg::FlushRemaningFrame( AVPacket* iVideoPacket )
{
	iVideoPacket->data = NULL;
	iVideoPacket->size = 0;
	int got_frame = 0;
	do
	{
		int ret = DecodeVideoPacket( iVideoPacket, &got_frame, 0, 0 );
		if( ret < 0 )
			break;
	} while( got_frame );


}

int
cLoaderFormatFFmpeg::DecodeVideoPacket( AVPacket* oPacket, int* oGotFrame, int* oFrame, cPixelBlockFF* oDst )
{
	*oGotFrame = 0;
	int decoded = oPacket->size;

	if( oPacket->stream_index != mVideoStreamId )
		return decoded;

	AVFrame *videoBufferFrame = av_frame_alloc();
	if( !videoBufferFrame )
	{
		return decoded;
	}

	int ret = avcodec_decode_video2( mVideoStream->codec, videoBufferFrame, oGotFrame, oPacket );
	if( ret < 0 )
	{
		av_frame_free( &videoBufferFrame );
		return ret;
	}

	if( !*oGotFrame )
	{
		av_frame_free( &videoBufferFrame );
		return decoded;
	}

	if( oFrame )
		//*oFrame = *oFrame + 1;
		*oFrame = PtsToFrame( videoBufferFrame->pkt_pts );

	if( !oDst )
	{
		av_frame_free( &videoBufferFrame );
		return decoded;
	}

	AVPicture src_picture = *(AVPicture *)videoBufferFrame;
	av_frame_free( &videoBufferFrame );

	AVPicture dst_picture;
	int error = avpicture_fill( &dst_picture, (const uint8_t*)oDst->Pixels(), AV_PIX_FMT_BGRA, oDst->Width(), oDst->Height() );
	if( error < 0 )
		return false;

	int height = sws_scale( mVideoSwsContext,
							src_picture.data, src_picture.linesize,
							0, mVideoStream->codec->height,
							dst_picture.data, dst_picture.linesize );
	if( height < 0 )
		return false;

	return decoded;
}




