/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE TVPAINT ANIMATION LOADER/SAVER PLUGINS  *
 * SOURCE CODE.  													*
 *																	*
 * This file uses some components of the FFmpeg project.			*
 * FFmpeg is distributed under the terms of the GNU Lesser General	*
 * Public License as published by the Free Software Foundation.		*
 *																	*
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS LIBRARY SOURCE IS     *
 * GOVERNED BY THE LGPL LICENSE. PLEASE READ THESE TERMS BEFORE 	*
 * DISTRIBUTING.													*
 *                                                                  *
 * THE TVPAINT ANIMATION LOADER/SAVER PLUGINS SOURCE CODE IS (C)    *
 * COPYRIGHT 2010-2015 BY THE TVPAINT DEVELOPPEMENT COMPANY         *
 * http://www.tvpaint.com/                                          *
 *                                                                  *
 ********************************************************************/

// ==================================================================
/** @file Format.LoaderFFmpeg.h
      @brief Header File: Format.LoaderFFmpeg
     @author Michael Schreiner
      $Date: $
        $Id: $
  @copyright (c) 2010-2015 TVPaint Developpement. All Rights Reserved.
*/
// ==================================================================

#ifndef __FORMAT_LOADERFFMPEG_HH
#define __FORMAT_LOADERFFMPEG_HH

#define __STDC_CONSTANT_MACROS  //Pour builder sur les anciens gcc ( pour linux )

#ifndef __FORMAT_BASELOADER_HH
#include "Format.BaseLoader.h"
#endif

extern "C"
{
	#include "libavcodec/avcodec.h"
	#include "libavformat/avformat.h"
	#include "libswscale/swscale.h"
	#include "libswresample/swresample.h"
	#include "libavutil/opt.h"
	#include "libavutil/mathematics.h"
}


struct cLoaderFormatFFmpegVideoHeader
{
	double  	mAspectRatio;
    int     	mColormap;
	double  	mDuration;
    AVRational 	mFrameRate;
    int     	mImageCount;
	int     	mVideoWidth;
	int     	mVideoHeight;
    eFieldOrder mFieldOrder;
};

struct cLoaderFormatFFmpegAudioHeader
{
	int mChannelCount;
    eChannelTypes mChannelType;
    int mBitsPerChannel;
    int mBytesPerSample;    // 16bits stereo(*2) => 4 bytes
    double mOffset;
    int mSampleCount;
    int mSamplingRate;
};

//---

class cLoaderFormatFFmpeg : public cLoaderFormat
{
public:
   ~cLoaderFormatFFmpeg();
	cLoaderFormatFFmpeg();

public:
	virtual bool Open( const char* iFilePath );
	virtual void Close();

	virtual const char* Version();

	virtual bool GetInfos( cLoaderFrameSequence* oFrameSequence );
	virtual bool GetInfos( cLoaderAudioSequence* oAudioSequence );
	virtual bool GetInfos( cLoaderAudioSequence* oAudioSequence, int iPreferedChannelsCount, int iPreferedBitsPerChannel, int iPreferedSamplingRate );

    virtual bool ReadFrame( int iFrameIndex, cPixelBlockFF* oDst );
    virtual bool ReadAudio( cAudioBlock* oDst );

private:
    //HEADER VIDEO
	uint16_t        Width( )          const { return  mVideoHeader.mVideoWidth; }
	uint16_t        Height( )         const { return  mVideoHeader.mVideoHeight; }
	eFieldOrder  	FieldOrder( )     const { return  mVideoHeader.mFieldOrder; }
	int             ImageCount( )     const { return  mVideoHeader.mImageCount; }
	AVRational      FrameRate( )      const { return  mVideoHeader.mFrameRate; }
	double          AspectRatio( )    const { return  mVideoHeader.mAspectRatio; }
	//uint8_t         Planes( )               { return  (mVideoHeader.mColormap & 0x07) + 1; }
	//double          Duration( )             { return  mVideoHeader.mDuration * mFormatContextPtr->streams[mVideoStreamId]->time_base.num / mFormatContextPtr->streams[mVideoStreamId]->time_base.den; }

    //HEADER AUDIO
	int             ChannelCount( )   const { return mAudioHeader.mChannelCount; }
	eChannelTypes   ChannelType( )    const { return mAudioHeader.mChannelType; }
	int             BitsPerChannel( ) const { return mAudioHeader.mBitsPerChannel;}
    int             BytesPerSample( ) const { return mAudioHeader.mBytesPerSample; }
    double          Offset( )         const { return mAudioHeader.mOffset; }
    int             SampleCount( )    const { return mAudioHeader.mSampleCount; }
    int             SamplingRate( )   const { return mAudioHeader.mSamplingRate; }

private:
	bool			OpenVideo();
	bool			OpenAudio();

	//Video
	bool 			LoadVideoByIndexSeeking( cPixelBlockFF* oDst, int iIdx );
	bool 			LoadVideoByTimeSeeking( cPixelBlockFF* oDst, double iTime );
	bool 			LoadVideo( cPixelBlockFF* oDst, int iIdx = -1, double iTime = 0.0 );

    //Audio
	bool 			LoadAudio( cAudioBlock* oDst );

	int64_t			FrameToPts( int iFrame ) const;
	int				PtsToFrame( int64_t iPts ) const;
	int				GetCodecDelay() const;

private:
	void			FlushRemaningFrame( AVPacket* iPacket );

	int 			DecodeVideoPacket( AVPacket* oPacket, int* oGotFrame, int* oFrame, cPixelBlockFF* oDst );
	int 			DecodeAudioPacket( AVPacket* oPacket, int* oGotFrame, cAudioBlock* oDst );

private:
    AVFormatContext*    mFormatContextPtr;

	int                 mVideoStreamId;
	AVStream*           mVideoStream;		// = mFormatContextPtr->streams[0]
	int                 mAudioStreamId;
	AVStream*           mAudioStream;		// = mFormatContextPtr->streams[1]

	//---

    cLoaderFormatFFmpegVideoHeader  mVideoHeader;

	AVCodecContext*     mVideoCodecContextPtr;
    AVCodec*            mVideoCodecPtr;
	int64_t				mVideoStartPTS;
	struct SwsContext*	mVideoSwsContext;

	//---

    cLoaderFormatFFmpegAudioHeader  mAudioHeader;

	AVCodecContext*     mAudioCodecContextPtr;
	AVCodec*            mAudioCodecPtr;

	struct SwrContext*	mAudioSwrContext;
    AVFrame*            mAudioBufferFrame;

	int					mAudioDstSamplesCountMax;
	uint8_t**			mAudioDstSamplesData;
	int					mAudioDstSamplesLinesize;
};


#endif	//__FORMAT_LOADERFFMPEG_HH

