/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE TVPAINT ANIMATION LOADER/SAVER PLUGINS  *
 * SOURCE CODE.  													*
 *																	*
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS LIBRARY SOURCE IS     *
 * GOVERNED BY A FREEBSD SOURCE LICENSE INCLUDED WITH THIS SOURCE   *
 * IN 'COPYING'. PLEASE READ THESE TERMS BEFORE DISTRIBUTING.       *
 *                                                                  *
 * THE TVPAINT ANIMATION LOADER/SAVER PLUGINS SOURCE CODE IS (C)    *
 * COPYRIGHT 2010-2015 BY THE TVPAINT DEVELOPPEMENT COMPANY         *
 * http://www.tvpaint.com/                                          *
 *                                                                  *
 ********************************************************************/

// ==================================================================
/** @file Format.SaverFFmpeg.h
      @brief Header File: Format.SaverFFmpeg
     @author Michael Scheiner & Eric SCHOLL
      $Date: $
        $Id: $
  @copyright (c) 2010-2015 TVPaint Developpement. All Rights Reserved.
*/
// ==================================================================

#ifndef __FORMAT_SAVERFFMPEG_HH
#define __FORMAT_SAVERFFMPEG_HH

#define __STDC_CONSTANT_MACROS  //Pour builder sur les anciens gcc ( pour linux )

#ifndef __FORMAT_BASESAVER_HH
#include "Format.BaseSaver.h"
#endif

#include <stdio.h>
#include <stdarg.h>

extern "C"
{
    #include "libavcodec/avcodec.h"
    #include "libavformat/avformat.h"
    #include "libswscale/swscale.h"
    #include "libswresample/swresample.h"
	#include "libavutil/opt.h"
}

class cSaverFormatFFmpeg : public cSaverFormat
{
public:
    enum eFormat
    {
        kFormatNULL = 0,
        kFormatQT = 1,
        kFormatMP4 = 2,
        kFormatMKV = 3,
        kFormatWebP = 4,
        kFormatWebM = 5,
    };

public:
	virtual  ~cSaverFormatFFmpeg();
	cSaverFormatFFmpeg( eFormat iFormat );

public:
    virtual const char* Copyright();
    virtual const char* Extension();
    virtual const char* FormatName();
    virtual const char* Options( int iIdx );
    virtual void Option( eSaverMode iIdx );
	virtual const char* Version();
    virtual bool UseBitrate();
    virtual int  Save(const char* iPathFile, cSaverFrameSequence* iFrameSequence, cSaverAudioSequence* iAudioSequence, int iOptionIdx, int iBitRate);
    virtual bool  Audio();
    virtual uint8_t SaverMode();
    virtual const char *Error();

private:
	virtual int					Bitrate( int iOption, int iMaxBitRate );

    virtual bool InitFormat( const char* iPathFile );
    bool SetOutputFormat( const char* format, const char* iPathFile, const char* shortName );
    virtual bool InitVideoStream();
    virtual bool InitAudioStream();
    virtual bool InitVideoCodecContext( AVCodecContext* ioCodecContext );
    virtual bool InitAudioCodecContext( AVCodecContext* ioCodecContext );

	virtual double				GetSamplesCountPerImage() const;
	virtual AVPixelFormat       GetPixelFormat( const AVCodec* iCodec );
	virtual AVSampleFormat      GetSampleFormat( const AVCodec* iCodec );
	virtual bool                InitVideoFrame();
	virtual bool                InitAudioFrame();
    virtual bool                WriteVideoFrame( cPixelBlockFF* iBlock, int iFrameNumber );
    virtual bool                WriteAudioFrame( int iFrameNumber );
	virtual bool                FlushFrames();
	virtual bool                FlushSamples();
    virtual void                CloseFFmpeg();
	void SaveErrorString(int code, const char* defaultMsg=NULL);
	void SaveErrorString(const char* error);

private:

    //TVPaint Tracks
    cSaverAudioSequence*	mSrcAudioTrack;
    cSaverFrameSequence*	mSrcVideoTrack;

    //TVpaint Options
    int 					mOption;
	eFormat                 mFormat;

    //Streams
	AVStream*               mVideoStream;		// = mFormatCtx->streams[0]
    AVStream*               mAudioStream;		// = mFormatCtx->streams[1]

    //Output
	AVOutputFormat*         mOutputFormat;		// = mFormatCtx->oformat
    AVFormatContext*        mFormatCtx;

    //Video
    AVFrame*                mVideoFrame;
	AVPicture				mVideoSrcPicture;
	AVPicture				mVideoDstPicture;
	struct SwsContext*		mVideoSwsContext;

    //Audio
	uint8_t**				mAudioSrcSamplesData;
	int						mAudioSrcSamplesLinesize;
	int						mAudioSrcSamplesCount;

	struct SwrContext*		mAudioSwrContext;
	AVFrame*				mAudioFrame;

	int						mAudioDstSamplesCountMax;
	uint8_t**				mAudioDstSamplesData;
	int						mAudioDstSamplesLinesize;
	int						mAudioDstSamplesSize;
	int						mAudioDstSamplesCount;

	int                     mFirstSample;
	int                     mStartSample;

    //Base
    int                     mNbPass;
    int                     mCurrentPass;
    int                     mCurrentFrame;

    char*					mLogBuffer;
    int						mLogBufferSize;

	char					mErrorMsg[256];
};



#endif // __FORMAT_SAVEROGG_HH


